#pragma once

#include "../debug/exception.h"
#include "../debug/exdebug.h"
#include "../utils/singleton.h"
#include "../threads/lock.h"
#include "../storage/storage.h"
#include "../utils/base64.h"
#include "../threads/thread.h"
#include "../event/base.h"

#include "exception.h"
#include "file_change_monitor.h"

#include <memory>
#include <unordered_map>
#include <mutex>
#include <string>
#include <cstdlib>
#include <atomic>
#include <list>

namespace resources
{
	/* manager class */
	class manager : public singleton<manager>
	{
		friend class singleton<manager>;

	private:

		/* loader thread class */
		class loader : public thread::base
		{
		public:

			// auto-start on ctor
			loader(manager& parent) : m_parent(parent)
			{
				start();
			}

		private:
			virtual void run(void) override
			{
				// loop until all resources are loaded
				while (this->m_parent.load_next());

				// sleep
				sleep(10e-3);
			}

		private:
			manager& m_parent;
		};

		/* bin base class */
		class base
		{
		public:
			virtual bool load(std::shared_ptr<void> p) = 0;
			virtual void reload(const std::string& name, bool async) = 0;
		};

	public:

		/* specialized bin class */
		template<class type> class bin : public base
		{
		public:

			/* cache class */
			class cache : public event::base
			{
				friend class bin<type>;

			public:

				// ctor
				cache(bin<type>& parent, const std::string& name) : m_parent(parent)
				{
					this->m_name = name;
				}

				// get name
				const std::string& name(void) const
				{
					return this->m_name;
				}

				// wait for event
				virtual bool wait(const double timeout = 0.0) const override
				{
					return this->m_available.wait(timeout);
				}

				// check if pending
				virtual bool is_pending(void) const override
				{
					return this->m_available.is_pending();
				}

				// get pointer
				std::shared_ptr<const type> get(void) const
				{
					// load pointer atomically
					auto p = std::atomic_load(&this->m_pointer);

					if (p != nullptr)
						return p;

					// default to null pointer if not set
					return this->m_parent.get_null_resource();
				}

				// reload resource
				void reload(bool async) const
				{
					this->m_parent.reload(name(), async);
				}

			private:

				// set pointer
				void set(std::shared_ptr<const type> p)
				{
					std::atomic_store(&this->m_pointer, p);

					// always trigger on set
					trigger();
				}

				// trigger event
				virtual void trigger(void) override
				{
					this->m_available.trigger();
				}

				// reset event
				virtual void reset(void) override
				{
					this->m_available.reset();
				}

			private:
				bin<type>& m_parent;

				std::string m_name;
				std::shared_ptr<const type> m_pointer;
				thread::event m_available;
			};

		public:

			bin(manager& parent) : m_parent(parent) {}

			// set null resource
			void set_null_resource(std::shared_ptr<const type> p)
			{
				std::atomic_store(&this->m_null_pointer, p);
			}

			// get resource
			std::shared_ptr<const cache> get(const std::string& rawname, bool async)
			{
				AUTOLOCK(this->m_lock);

				// build resource name
				auto resname = format_name(rawname);

				// check if it already exists
				auto it = this->m_list.find(resname);

				if (it != this->m_list.end())
				{
					auto p = it->second.lock();

					// return if still valid
					if (p != nullptr)
						return p;

					_debug("reloading resource %s", resname.c_str());

					// otherelse recreate one
					p = create(resname, async);

					it->second = p;

					return p;
				}

				_debug("creating resource %s", resname.c_str());

				// otherelse create
				auto p = create(resname, async);

				// add to list
				this->m_list.emplace(std::make_pair(resname, p));

				// return pointer
				return p;
			}

			// load resource
			virtual bool load(std::shared_ptr<void> p_raw) override
			{
				// cast
				std::shared_ptr<cache> p_cache = std::static_pointer_cast<cache>(p_raw);

				if (__ISNULLPTR(p_cache))
					return false;

				// create type
				_debug("loading resource %s", p_cache->name().c_str());

				try
				{
					// assign new
					p_cache->set(create_object(p_cache->name()));

					// return successfull
					return true;
				}
				catch (exception::base& err)
				{
					_error("resource loading triggered exception \"%s\" !", err.to_string().c_str());
				}
				catch (std::exception& err)
				{
					_error("resource loading triggered exception \"%s\" !", err.what());
				}
				catch (...)
				{
					_error("resource loading triggered unknown exception !");
				}

				// set to null in case of error (default to null resource)
				p_cache->set(nullptr);

				p_cache->trigger();

				// return unsuccessfull
				return false;
			}

			// reload resource
			virtual void reload(const std::string& rawname, bool async) override
			{
				// format
				auto resname = format_name(rawname);

				_debug("reloading resource %s...", resname.c_str());

				// find resource if existing
				AUTOLOCK(this->m_lock);

				auto it = this->m_list.find(resname);

				if (it == this->m_list.end())
				{
					_warning("resource %s not found", resname.c_str());

					return;
				}

				auto p = it->second.lock();

				// reload resource only if there is an active one
				if (p == nullptr)
					return;

				// reset trigger
				p->reset();

				// defer for later
				if (async)
					this->m_parent.add_pending(typeid(type).name(), p);
				// create now
				else
					p->set(create_object(resname));
			}

			// null pointer
			std::shared_ptr<const type> get_null_resource(void) const
			{
				return std::atomic_load(&this->m_null_pointer);
			}

			// cleanup
			bool cleanup(const std::string& rawname)
			{
				auto resname = format_name(rawname);

				_debug("cleaning up %s", resname.c_str());

				AUTOLOCK(this->m_lock);

				// find resource
				auto it = this->m_list.find(resname);

				if (it == this->m_list.end())
				{
					_warning("resource %s not found", resname.c_str());

					return false;
				}

				// skip if still active
				if (!it->second.expired())
					return false;

				// unmonitor resource
				static const char* _file = "file://";

				if (str_begins_with(resname, _file))
					this->m_parent.unmonitor(resname.substr(strlen(_file)));

				// remove from list
				this->m_list.erase(it);

				// return successful
				return true;
			}

		private:

			// format resource name
			static std::string format_name(const std::string& name)
			{
				static const char* _file = "file://";

				// decode name for files
				if (str_begins_with(name, _file))
				{
					auto filename = name.substr(strlen(_file));
					auto s = split_file_path(filename);

					return _file + build_path({ s.directory, s.file });
				}

				// otherelse just return original name
				return name;
			}

			// create cache structure
			std::shared_ptr<cache> create(const std::string& resname, bool async)
			{
				// create cache object
				auto p = std::make_shared<cache>(*this, resname);

				// defer for later
				if (async)
					this->m_parent.add_pending(typeid(type).name(), p);
				// create now
				else
					p->set(create_object(resname, true));

				// return pointer
				return p;
			}

			// create object
			std::shared_ptr<const type> create_object(const std::string& resname, bool unmonitor_on_failure=false)
			{
				static const char* _file = "file://";

				// special case for files
				if (str_begins_with(resname, _file))
				{
					auto filename = resname.substr(strlen(_file));

					// monitor file even if instanciation does not succeed
					this->m_parent.monitor(filename);

					// create pointer
					try
					{
						auto p = std::make_shared<const type>(storage::load_buffer_from_file(filename));

						// return pointer
						return p;
					}
					catch (...)
					{
						if (unmonitor_on_failure)
							this->m_parent.unmonitor(filename);

						throw;
					}
				}

				// split type://handle
				auto pos = resname.find("://");

				// throw exception if not found
				if(pos == std::string::npos)
					throw_exception(ill_formed_resource_type_exception, resname);

				// split
				auto restype = resname.substr(0, pos);
				auto handle = resname.substr(pos + 3);

				// load
				return std::make_shared<const type>(this->m_parent.load(restype, handle));
			}

		private:
			manager& m_parent;

			std::shared_ptr<const type> m_null_pointer;

			std::mutex m_lock;
			std::unordered_map<std::string, std::weak_ptr<cache>> m_list;
		};

	public:

		/* auto_register_load_type class */
		class auto_register_load_type
		{
		public:
			auto_register_load_type(const std::string& type, std::function<storage::buffer(const std::string&)> func)
			{
				::get_instance<manager>()->register_loader_type(type, func);
			}
		};

	public:

		template<class type> using cache_t = typename bin<type>::cache;

		// cleanup everything on dtor
		~manager(void)
		{
			try
			{
				if (this->m_loader != nullptr)
					this->m_loader->stop();

				this->m_loader.reset();

				this->m_change_monitor.reset();

				std::lock_guard<std::recursive_mutex> lock(this->m_lock);

				this->m_pending.clear();
				this->m_bins.clear();
			}
			catch (...) {}
		}

		// get resources
		template<class type> std::shared_ptr<bin<type>> get(void)
		{
			auto type_name = typeid(type).name();

			// find bin
			std::lock_guard<std::recursive_mutex> lock(this->m_lock);

			auto it = this->m_bins.find(type_name);

			// if existing
			if (it != this->m_bins.end())
				return std::static_pointer_cast<bin<type>>(it->second);

			// if not, create it
			auto p = std::make_shared<bin<type>>(*this);

			this->m_bins.emplace(type_name, p);

			return p;
		}

		// reload
		void reload(const std::string& filename)
		{
			std::lock_guard<std::recursive_mutex> lock(this->m_lock);

			for (auto& v : this->m_bins)
				if (v.second != nullptr)
					v.second->reload(filename, true);
		}

		// count pending resources
		size_t num_pending(void) const
		{
			std::lock_guard<std::recursive_mutex> lock(this->m_lock);

			return this->m_pending.size();
		}

		// count watched resources
		size_t num_watched(void) const
		{
			if (this->m_change_monitor == nullptr)
				return 0;

			return this->m_change_monitor->num_watched();
		}

	private:

		// ctor
		manager(void)
		{
			this->m_loader = std::make_shared<loader>(*this);
			this->m_change_monitor = std::make_shared<file_change_monitor>(std::bind(&manager::on_change, this, std::placeholders::_1));
		}

		// add to change monitor
		bool monitor(const std::string& filename)
		{
			if (this->m_change_monitor == nullptr)
				return false;

			return this->m_change_monitor->add(filename);
		}

		// remove from change monitor
		bool unmonitor(const std::string& filename)
		{
			if (this->m_change_monitor == nullptr)
				return false;

			return this->m_change_monitor->remove(filename);
		}

		// file has been changed
		void on_change(const std::string& filename)
		{
			_debug("file \"%s\" has been changed", filename.c_str());

			// schedule reload
			reload("file://" + filename);
		}

		// add pending resource
		void add_pending(const std::string& type, std::shared_ptr<void> p)
		{
			_debug("adding deferred loading for resource type %s", type.c_str());

			{
				std::lock_guard<std::recursive_mutex> lock(this->m_lock);

				this->m_pending.push_back(std::make_pair(type, p));
			}
		}

		// load next pending resource
		bool load_next(void)
		{
			std::lock_guard<std::recursive_mutex> lock(this->m_lock);

			std::string bintype;
			std::shared_ptr<void> p_cache;

			{
				// no more resources to load
				if (this->m_pending.empty())
					return false;

				// get next resources
				auto curr = this->m_pending.front();
				this->m_pending.pop_front();

				bintype = curr.first;
				p_cache = curr.second;
			}

			std::shared_ptr<base> p_bin;

			{
				auto it = this->m_bins.find(bintype);

				// should never happend
				if (it == this->m_bins.end())
					throw_exception(unknown_resource_type_exception);

				p_bin = it->second;
			}

			// load resource
			p_bin->load(p_cache);

			// return successful
			return true;
		}

		// register a resource loader type
		void register_loader_type(const std::string& type, std::function<storage::buffer(const std::string&)> func)
		{
			std::lock_guard<std::recursive_mutex> lock(this->m_lock);

			// discard invalid
			if (!func)
				throw_exception(invalid_loader_type_exception, type);

			// check if type already exists
			if (this->m_loader_funcs.find(type) != this->m_loader_funcs.end())
				throw_exception(loader_type_already_exists_exception, type);

			// otherelse insert
			this->m_loader_funcs.emplace(std::make_pair(type, func));
		}

		// load a given type
		storage::buffer load(const std::string& type, const std::string& handle)
		{
			std::lock_guard<std::recursive_mutex> lock(this->m_lock);

			// find type
			auto it = this->m_loader_funcs.find(type);

			if (it == this->m_loader_funcs.end())
				throw_exception(unknown_loader_type_exception, type);

			// check if valid
			if (!it->second)
				throw_exception(invalid_loader_type_exception, type);

			// instantiate
			return it->second(handle);
		}

	private:
		std::shared_ptr<loader> m_loader;
		std::shared_ptr<file_change_monitor> m_change_monitor;

		mutable std::recursive_mutex m_lock;

		std::list<std::pair<std::string, std::shared_ptr<void>>> m_pending;
		std::unordered_map<std::string, std::shared_ptr<base>> m_bins;

		std::unordered_map<std::string, std::function<storage::buffer(const std::string&)>> m_loader_funcs;
	};

	// retrieve cache pointer
	template<class type> std::shared_ptr<const manager::cache_t<type>> get_cache_pointer(const std::string& name, bool async)
	{
		// get instance
		auto p = get_instance<manager>()->get<type>();

		// shall never happen
		if (__ISNULLPTR(p))
			throw_exception(unknown_resource_type_exception);

		// get cache pointer
		return p->get(name, async);
	}

	// retrieve cache pointer
	template<class type> bool cleanup_cache_pointer(const std::string& name)
	{
		// get instance
		auto p = get_instance<manager>()->get<type>();

		// shall never happen
		if (__ISNULLPTR(p))
			throw_exception(unknown_resource_type_exception);

		// cleanup cache pointer
		return p->cleanup(name);
	}

	// set null resource
	template<class type> void set_null_resource(std::shared_ptr<const type> p_null)
	{
		// get instance
		auto p = get_instance<manager>()->get<type>();

		// shall never happen
		if (__ISNULLPTR(p))
			throw_exception(unknown_resource_type_exception);

		// set null resource
		p->set_null_resource(p_null);
	}

	// get null resource
	template<class type> std::shared_ptr<const type> get_null_resource(void)
	{
		// get instance
		auto p = get_instance<manager>()->get<type>();

		// shall never happen
		if (__ISNULLPTR(p))
			throw_exception(unknown_resource_type_exception);

		// set null resource
		return p->get_null_resource();
	}
};

using resources_manager_t = resources::manager;

#define REGISTER_RESOURCE_LOADER_EX(func, type, name)	resources::manager::auto_register_load_type g_register_resources_load_##name(type, &func);
#define REGISTER_RESOURCE_LOADER(func, type)			REGISTER_RESOURCE_LOADER_EX(func, type, func);