#include <stdio.h>

#include "../core/utils/utils.h"
#include "../core/debug/evemon.h"
#include "../core/debug/exception.h"
#include "../core/threads/event.h"
#include "../core/net/net.h"
#include "../core/instancers/instancer.h"
#include "../core/storage/storage.h"
#include "../core/test/test.h"
#include "../core/resources/resources.h"

#include "intval.h"

#include <filesystem>

/* Reload Test */
class reload_test : public test::base
{
public:
	virtual void run(void) const override
	{
		_comment("Resources Reload Test");

		// auto reload
		{
			const char* resname = "file://test.txt";

			// write file
			{
				FILE* p = nullptr;

				fopen_s(&p, "test.txt", "wb+");

				if (p == nullptr)
					throw_exception(cannot_open_file_exception, "test.txt");

				fputs("10", p);

				fclose(p);
			}

			// load resource
			resources::proxy<intval> r(resname);

			// check that it was correctly constructed
			EXPECT_NOTNULL("Resource shall not be null", r.get());
			EXPECT_INT("Resource value shall be 10", r->get(), 10);

			// rewrite file
			{
				// wait a bit for timestamp
				std::this_thread::sleep_for(std::chrono::duration<double, std::ratio<1>>(0.1));

				FILE* p = nullptr;

				fopen_s(&p, "test.txt", "wb+");

				if (p == nullptr)
					throw_exception(cannot_open_file_exception, "test.txt");

				fputs("20", p);

				fclose(p);
			}

			// wait for reload
			thread::event e;

			bool status = e.do_until([&](void)
							{
								try
								{
									if (r->get() == 20)
										e.trigger();
								}
								catch (...) {}

							}, 1.0);

			// check that it was correctly constructed
			EXPECT_TRUE("Resource auto-reload shall not timeout", status);
			EXPECT_NOTNULL("Resource after auto-reload shall not be null", r.get());
			EXPECT_INT("Resource value after auto-reload shall now be 20", r->get(), 20);
		}

		// removing resource reload
		{
			const char* resname = "file://test.txt";

			// write file
			{
				FILE* p = nullptr;

				fopen_s(&p, "test.txt", "wb+");

				if (p == nullptr)
					throw_exception(cannot_open_file_exception, "test.txt");

				fputs("10", p);

				fclose(p);
			}

			// load resource
			resources::proxy<intval> r(resname);

			// check that it was correctly constructed
			EXPECT_NOTNULL("Resource shall not be null", r.get());
			EXPECT_INT("Resource value shall be 10", r->get(), 10);

			// set null resource
			resources::set_null_resource(std::make_shared<const intval>(-1));

			// remove resource
			std::filesystem::remove("test.txt");

			// wait for reload
			thread::event e;

			bool status = e.do_until([&](void)
				{
					try
					{
						if (r->get() == -1)
							e.trigger();
					}
					catch (...) {}

				}, 1.0);

			// check that it was correctly constructed
			EXPECT_TRUE("Resource auto-reload after removal shall not timeout", status);
			EXPECT_NOTNULL("Resource auto-reload after removal shall not be null", r.get());
			EXPECT_INT("Resource value after auto-reload removal shall now be -1", r->get(), -1);

			// clear null resource
			resources::set_null_resource<intval>(nullptr);
		}

		// reload invalid resource
		{
			const char* resname = "file://test.txt";

			// remove first
			std::filesystem::remove("test.txt");

			// load resource
			resources::async_proxy<intval> r(resname);

			// check that it was correctly constructed
			EXPECT_NULL("Resource shall be null at first", r.get());

			// write file
			{
				FILE* p = nullptr;

				fopen_s(&p, "test.txt", "wb+");

				if (p == nullptr)
					throw_exception(cannot_open_file_exception, "test.txt");

				fputs("10", p);

				fclose(p);
			}

			// wait for reload
			thread::event e;

			bool status = e.do_until([&](void)
				{
					try
					{
						if (r->get() == 10)
							e.trigger();
					}
					catch (...) {}

				}, 1.0);

			// check that it was correctly constructed
			EXPECT_TRUE("Resource auto-reload after initial write shall not timeout", status);
			EXPECT_NOTNULL("Resource auto-reload after initial write shall not be null", r.get());
			EXPECT_INT("Resource value after auto-reload initial write shall now be 10", r->get(), 10);
		}

		// force reload
		{
			const char* resname = "file://test.txt";

			// write file
			{
				FILE* p = nullptr;

				fopen_s(&p, "test.txt", "wb+");

				if (p == nullptr)
					throw_exception(cannot_open_file_exception, "test.txt");

				fputs("10", p);

				fclose(p);
			}

			// load resource
			resources::proxy<intval> r(resname);

			// check that it was correctly constructed
			EXPECT_NOTNULL("Resource shall not be null", r.get());
			EXPECT_INT("Resource value shall be 10", r->get(), 10);

			// rewrite file
			{
				FILE* p = nullptr;

				fopen_s(&p, "test.txt", "wb+");

				if (p == nullptr)
					throw_exception(cannot_open_file_exception, "test.txt");

				fputs("20", p);

				fclose(p);
			}

			// reload
			r.reload();

			// check that it was correctly constructed
			EXPECT_NOTNULL("Resource after manual reload shall not be null", r.get());
			EXPECT_INT("Resource value after manual reload shall now be 20", r->get(), 20);

			// HACK: wait for resource loading termination, should not occur in real life
			// due to resource being modified just before being destroyed and then recreated
			// the reload can then occur just after the resource has been recreated but before
			// it is accessed!
			std::this_thread::sleep_for(std::chrono::duration<double, std::ratio<1>>(0.1));

			EXPECT_TRUE("Resource loading done before timeout", r.wait(1.0));
		}
	}
};

REGISTER_TEST(reload_test);